### This code was developed for the research study that led to the paper [1],
### within a partnership between the Scuola Normale Superiore and LIST (http://www.list-group.com).
### It is released with no guarantee and for research purposes only.
### It includes the functions for estimating and simulating the models
### proposed in the paper. A syntactical description of the functions
### for a correct usage is provided in a separate file.
### Researchers using this code for their own research are invited to cite the paper [1].
### [1] G. Bormetti, L. M. Calcagnile, M. Treccani, F. Corsi, S. Marmi, F. Lillo, Modelling systemic price cojumps with Hawkes factor models, Quantitative Finance 15 (7), 1137-1156


########## Cojumps and Hawkes ##########
# general scheme

#MC <- 200
#Poisson_MJ_CJ(asset_list,prefixes,param,path,totlen,windowlengths)	# ok
#Hawkes_1D_MJ(asset_list,prefixes,param,path,totlen,windowlengths,MC)	# ok
#Hawkes_2D_CJ(asset_list,prefixes,param,path,totlen,windowlengths,MC)
#Hawkes_1F_MJ_CJ(asset_list,prefixes,param,path,totlen,windowlengths)

########## End Cojumps and Hawkes ##########


##### begin of Poisson_MJ_CJ

Poisson_MJ_CJ <- function(asset_list,prefixes,param,path,totlen,windowlengths) {
	# Count and save MJ and CJ.
	print("Counting and saving observed MJ's and CJ's...")
	count_and_save_MJ_CJ(asset_list,prefixes,param,path,windowlengths)

	# Plot empirical MJ against Poisson theoretical values.
	print("Plotting MJ's against Poisson theoretical values...")
	plot_MJ_against_Poisson(asset_list,prefixes,param,path,totlen,windowlengths)

	# Checking whether the pairs are independent Poisson is meaningless, since the single-asset jump series are usually not Poisson themselves.
	#print("Plotting CJ's against Poisson theoretical values...")
	#plot_CJ_against_Poisson(asset_list,prefixes,param,path,totlen,windowlengths)
}


count_and_save_MJ_CJ <- function(asset_list,prefixes,param,path,windowlengths) {
	for (i in 1:length(asset_list)) {
		# Look for multiple jumps.
		if (i %% 10 == 0) print(paste("asset",i))
		s1 <- loadjumps(asset_list[i],prefixes,param,path)	# xts with TRUE (jump), FALSE, or NA.
		windex1 <- which(s1 == TRUE)

		MJprob <- rep(0,length(windowlengths))
		names(MJprob) <- as.character(windowlengths)
		MJprob <- count_MJ(windex1,windowlengths,length(s1))
		write(c(i,i,MJprob),file=paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs"),append=TRUE,ncolumns=length(MJprob)+2)

		for (j in (i+1):length(asset_list)) {
			if ((j <= length(asset_list)) & (j > i)) {
				#print(paste(i,j))
				s2 <- loadjumps(asset_list[j],prefixes,param,path)
				windex2 <- which(s2 == TRUE)

				CJprob <- rep(0,length(windowlengths))
				names(CJprob) <- as.character(windowlengths)
				CJprob <- count_CJ(windex1,windex2,windowlengths,length(s2))
				write(c(i,j,CJprob),file=paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs"),append=TRUE,ncolumns=length(CJprob)+2)
				write(c(j,i,CJprob),file=paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[j],"_obs"),append=TRUE,ncolumns=length(CJprob)+2)
			}
		}
	}
}


loadjumps <- function(assetname,prefixes,param,path,sign="all") {
	ju <- readRDS(paste0(path$jumps,prefixes[period_counter],"_",assetname,"_r_over_sigma.RDS"))
	jumpthres <- param$jump_thres
	jumpos <- ju > jumpthres
	jumneg <- ju < -jumpthres
	if (sign == "all") return(jumpos | jumneg)		# Note that (NA | TRUE) = (TRUE | NA) = TRUE.
	else if (sign == "pos") return(jumpos)
	else if (sign == "neg") return(jumneg)
}


count_MJ <- function(wind,wlen,len) {
	mj <- rep(0,length(wlen))
	names(mj) <- as.character(wlen)
	wind <- unique(wind)	# To avoid repeated indices due to discretisation of a continuous process.
	for (wl in wlen) {
		w <- wind %/% wl + 1
		t <- table(w)
		mj[as.character(wl)] <- length(t[t>1]) / (len %/% wl)
	}
	mj
}


count_CJ <- function(wind1,wind2,wlen,len) {
	cj <- rep(0,length(wlen))
	names(cj) <- as.character(wlen)
	for (wl in wlen) {
		w1 <- wind1 %/% wl + 1
		w2 <- wind2 %/% wl + 1
		cj[as.character(wl)] <- length(intersect(w1,w2)) / (len %/% wl)
	}
	cj
}


plot_MJ_against_Poisson <- function(asset_list,prefixes,param,path,totlen,windowlengths) {
	nwlen <- length(windowlengths)
	alpha1 <- 0.01
	alpha5 <- 0.05
	alpha1Bonf <- alpha1 / nwlen
	alpha5Bonf <- alpha5 / nwlen
	nsd1 <- qnorm(1 - alpha1Bonf)
	nsd5 <- qnorm(1 - alpha5Bonf)

	pdf(paste0(path$cojumps,"Poisson/",prefixes[period_counter],"_MJ_Pois.pdf"),width=8,height=6)
		# WARNING: pay attention to the Bonferroni correction.
		for (i in 1:length(asset_list)) {
			s <- loadjumps(asset_list[i],prefixes,param,path)
			nju <- length(which(s == TRUE))
			lambda <- nju / totlen
			thmean <- 1 - (1-lambda)^windowlengths - windowlengths*lambda*(1-lambda)^(windowlengths-1)
			thsd <- theoretical_sd_MJ(windowlengths,lambda,totlen)

			MJp <- read.table(paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs"))[i,][-(1:2)]
			plot(windowlengths,thmean,type="o",col="lightblue",ylim=c(0,max(MJp)),ylab="MJ frequency",xlab="window width",main=paste("MJ probability of ",asset_list[i]))

			tmp <- thmean + nsd1*thsd; tmp1 <- 0.5 * (1 + sign(tmp)) * tmp
			lines(windowlengths,tmp1,type="o",col="red")
			tmp <- thmean - nsd1*thsd; tmp1 <- 0.5 * (1 + sign(tmp)) * tmp
			lines(windowlengths,tmp1,type="o",col="red")

			tmp <- thmean + nsd5*thsd; tmp1 <- 0.5 * (1 + sign(tmp)) * tmp
			lines(windowlengths,tmp1,type="o",col="orange")
			tmp <- thmean - nsd5*thsd; tmp1 <- 0.5 * (1 + sign(tmp)) * tmp
			lines(windowlengths,tmp1,type="o",col="orange")

			tmp <- as.numeric(MJp[1:nwlen])
			lines(windowlengths,tmp,type="o",col="blue")
			legend(x="topleft",legend=c("observed","Poisson theoretical mean","Poisson theoretical 99% c.l.","Poisson theoretical 95% c.l."),col=c("blue","lightblue","red","orange"),lty="solid",pch="o",inset=0.04)
		}
	dev.off()
}


plot_CJ_against_Poisson <- function(asset_list,prefixes,param,path,totlen,windowlengths) {
	nwlen <- length(windowlengths)
	alpha1 <- 0.01
	alpha1Bonf <- alpha1 / nwlen
	nsd1 <- qnorm(1 - alpha1Bonf)

	pdf(paste0(path$cojumps,"Poisson/",prefixes[period_counter],"_CJ_Pois.pdf"),width=8,height=6)
		for (i in 1:(length(asset_list)-1)) {
			si <- loadjumps(asset_list[i],prefixes,param,path)
			njui <- length(which(si == TRUE))
			lambdai <- njui / totlen
			for (j in (i+1):length(asset_list)) {
				sj <- loadjumps(asset_list[j],prefixes,param,path)
				njuj <- length(which(sj == TRUE))
				lambdaj <- njuj / totlen
				thmean <- (1 - (1-lambdai)^windowlengths) * (1 - (1-lambdaj)^windowlengths)
				thsd <- theoretical_sd_CJ(windowlengths,lambdai,lambdaj,totlen)

				CJp <- read.table(paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs"))[j,][-(1:2)]
				plot(windowlengths,thmean,type="o",col="lightblue",ylim=c(0,max(CJp)),ylab="CJ frequency",xlab="window width",main=paste("CJ probability of",asset_list[i],"and",asset_list[j]))

				tmp <- thmean + nsd1*thsd; tmp1 <- 0.5 * (1 + sign(tmp)) * tmp
				lines(windowlengths,tmp1,type="o",col="red")
				tmp <- thmean - nsd1*thsd; tmp1 <- 0.5 * (1 + sign(tmp)) * tmp
				lines(windowlengths,tmp1,type="o",col="red")

				tmp <- as.numeric(CJp[1:nwlen])
				lines(windowlengths,tmp,type="o",col="blue")
				legend(x="topleft",legend=c("observed","Poisson theoretical mean","Poisson theoretical 99% c.l."),col=c("blue","lightblue","red"),lty="solid",pch="o",inset=0.04)
			}
		}
	dev.off()
}


theoretical_sd_MJ <- function(wl,lamb,len) {
	pl <- 1 - (1-lamb)^wl - wl*lamb*(1-lamb)^(wl-1)
	n <- len %/% wl
	temp <- abs((pl - pl^2) / n)
	sqrt(temp)
}


theoretical_sd_CJ <- function(wl,lamb1,lamb2,len) {
	pl <- (1 - (1-lamb1)^wl) * (1 - (1-lamb2)^wl)
	n <- len %/% wl
	temp <- abs((pl - pl^2) / n)
	sqrt(temp)
}

##### end of Poisson_MJ_CJ

##### start of Hawkes_1D_MJ

Hawkes_1D_MJ <- function(asset_list,prefixes,param,path,totlen,windowlengths,MC) {
	# Calibrate 1D Hawkes processes and perform MC simulations.
	print("Estimating and simulating 1D Hawkes...")
	calibrate_and_simulate_1DHawkes(asset_list,prefixes,param,path,totlen,windowlengths,MC)

	# Plot MJ against 1DHawkes MC simulation results.
	print("Plotting MJ's against 1DHawkes MC...")
	plot_MJ_against_1DHawkes(asset_list,prefixes,path,windowlengths)
}


calibrate_and_simulate_1DHawkes <- function(asset_list,prefixes,param,path,totlen,windowlengths,MC) {
	nwlen <- length(windowlengths) 

	for (i in 1:length(asset_list)) {
		if (i %% 10 == 0) print(i)
		# Look for MJs.
		s <- loadjumps(asset_list[i],prefixes,param,path)
		windex <- which(s == TRUE)
		waiting_times <- diff(windex)
		lambda_init <- 1 / mean(waiting_times)
		alpha_init  <- 0.015
		beta_init   <- 0.2
		mleHawkes <- maxLik(logLik=logLikFun_Hawkes,start=c(lambda_init,alpha_init,beta_init),windex=windex)
		Hparams <- summary(mleHawkes)$estimate
		Hestim  <- c(Hparams[,1],Hparams[,4])
		write(Hestim,file=paste0(path$cojumps,"Hawkes_1D_MJ/",prefixes[period_counter],"_1DH_estim_and_pval"),append=TRUE,ncolumns=6)

		MJprob <- array(c(0),dim=c(MC,nwlen))
		for (mc in 1:MC) {
			#if (mc %% 50 == 0) print(mc)
			newHawkes <- simHawkes(totlen,mleHawkes$estimate)
			MJprob[mc,] <- count_MJ(newHawkes,windowlengths,totlen)
		}

		mnmc <- colMeans(MJprob)
		lw99qmc <- rep(0,nwlen)
		up99qmc <- rep(0,nwlen)
		lw95qmc <- rep(0,nwlen)
		up95qmc <- rep(0,nwlen)

		for (j in 1:nwlen) {
			tmp <- quantile(MJprob[,j],probs=c(0.005/nwlen,1-0.005/nwlen))
			lw99qmc[j] <- tmp[1]
			up99qmc[j] <- tmp[2]
			tmp <- quantile(MJprob[,j],probs=c(0.025/nwlen,1-0.025/nwlen))
			lw95qmc[j] <- tmp[1]
			up95qmc[j] <- tmp[2]
		}
		outputfile <- paste0(path$cojumps,"Hawkes_1D_MJ/",prefixes[period_counter],"_",asset_list[i],"_1DH")
		write(c(i,i,mnmc)   ,file=outputfile,append=TRUE,ncolumns=nwlen+2)
		write(c(i,i,lw99qmc),file=outputfile,append=TRUE,ncolumns=nwlen+2)
		write(c(i,i,up99qmc),file=outputfile,append=TRUE,ncolumns=nwlen+2)
		write(c(i,i,lw95qmc),file=outputfile,append=TRUE,ncolumns=nwlen+2)
		write(c(i,i,up95qmc),file=outputfile,append=TRUE,ncolumns=nwlen+2)
	}
}


HawkesTotIntensity <- function(t_events,pars) {
	# HawkesTotIntensity computes the total integrated intensity in
	# [0,t_n] of a 1-dim Hawkes process with last event occurring at time t_n.

	lmb <- pars[1]
	aph <- pars[2]
	bet <- pars[3]

	n <- length(t_events)
	A <- rep(0,(n-1))
	wtimes <- diff(t_events)

	A[1] <- 1  
	for (i in 2:(n-1)) {  # ricursive computation of A
		A[i] <- 1 + exp(-bet * (wtimes[i-1])) * A[i-1]
	}

	# Computation of the total integrated intensity in [0,t_n]
	TotLambda <- sum(c(lmb * t_events[1], wtimes * lmb + aph / bet * (1 - exp(-bet * wtimes)) * A))

	TotLambda
}


logLikFun_Hawkes <- function(pars,windex) {
	lmb <- pars[1]
	aph <- pars[2]
	bet <- pars[3]

	n <- length(windex)
	R <- rep(0,n)
	for (k in 2:n) {  # ricursive computation of R
		R[k] <- exp(-bet * (windex[k] - windex[k-1])) * (1 + R[k-1])
	}

	# Computation of the log-likelihood
	logL <- windex[n] - HawkesTotIntensity(windex,pars) + sum(log(lmb + aph * R))

	logL 
}


lambda.fun <- function(pars,t_events,t) {
	# This function computes the intensity lmb(t) of a 1-dim Hawkes process at time t.

	lmb <- pars[1]
	aph <- pars[2]
	bet <- pars[3]

	past_events <- t_events[t_events < t]
	l <- lmb + aph * sum(exp(-bet * (t - past_events)))
	l
}


simHawkes <- function(T,pars) { 
	# simHawkes simulates a one-dimensional Hawkes process with P=1 and
	# parameters lmb, alpha and bet given in input, on the time interval [0,T].
	# INPUT:
	# T = final time
	# pars = c(lmb, alpha, bet)
	# OUTPUT:
	# t_events = vector of event times of the process;

	lmb <- pars[1]
	aph <- pars[2]

	lmb_max  <- rep(0,12 * T)
	tstar    <- rep(0,12 * T)
	t_events <- rep(0,10 * T)

	i <- 1
	istar <- 1
	lstar <- lmb
	tstar[istar] <- 0
	lmb_max[istar] <- lmb
	U <- runif(1) 
	s <- - 1 / lstar * log(U)  # generate the 1st event time from a Exp(lstar) dist.

	# General routine
	while(s <= T) {
		t_events[i] <- s # event time
		i <- i+1
		istar <- istar + 1
		lstar <- lambda.fun(pars,t_events[1:(i-1)],s) + aph
		tstar[istar] <- s
		lmb_max[istar] <- lstar
		U <- runif(1)
		s <- s - 1 / lstar * log(U)
		while(s <= T) {
			D <- runif(1) 
			if (D > lambda.fun(pars,t_events[1:(i-1)],s) / lstar) {
				istar <- istar + 1
				lstar <- lambda.fun(pars,t_events[1:(i-1)],s)
				tstar[istar] <- s
				lmb_max[istar] <- lstar
				U <- runif(1)
				s <- s - 1 / lstar * log(U)
			} else {
				break # exit the inner while and go to outer
			}
		}
	}

	n_events <- i - 1
	tstar <- tstar[1:istar]
	lmb_max <- lmb_max[1:istar]
	# rounding 
	t_events <- round(t_events[1:n_events])
	t_events 
}


plot_MJ_against_1DHawkes <- function(asset_list,prefixes,path,windowlengths) {
	pdf(paste0(path$cojumps,"Hawkes_1D_MJ/",prefixes[period_counter],"_MJ_1DH.pdf"),width=8,height=6)
		for (i in 1:length(asset_list)) {
			MJp <- read.table(paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs"))[i,][-(1:2)]
			MCres <- read.table(paste0(path$cojumps,"Hawkes_1D_MJ/",prefixes[period_counter],"_",asset_list[i],"_1DH"))[1:5,-(1:2)]
			plotMJCJ_withMCcl(windowlengths,MJp,MCres,ylab="MJ frequency",plottitle=paste("MJ probability of",asset_list[i]))
		}
	dev.off()
}

##### end of Hawkes_1D_MJ

##### start of Hawkes_2D_CJ

Hawkes_2D_CJ <- function(asset_list,prefixes,param,path,totlen,windowlengths,MC) {
	# Calibrate 2D Hawkes processes and perform MC simulations.
	print("Estimating and simulating 2D Hawkes...")
	calibrate_and_simulate_2DHawkes(asset_list,prefixes,param,path,totlen,windowlengths,MC)
	
	# Plot CJ against 2DHawkes MC simulation results.
	print("Plotting CJ's against 2DHawkes MC...")
	plot_CJ_against_2DHawkes(asset_list,prefixes,path,windowlengths)

	# Plot MJ against 2DHawkes MC simulation results.
	print("Plotting MJ's against 2DHawkes MC...")
	plot_MJ_against_2DHawkes(asset_list,prefixes,path,windowlengths)
}


calibrate_and_simulate_2DHawkes <- function(asset_list,prefixes,param,path,totlen,windowlengths,MC) {
	nwlen <- length(windowlengths) 
	maxsann <- 100	# Maximum number of optimisation attempts with the simulated annealing method.

	alpha_init <- 0.015
	beta_init  <- 0.2

	for (i in 1:length(asset_list)) {
		# Look for CJs.
		si <- loadjumps(asset_list[i],prefixes,param,path)
		windexi <- which(si == TRUE)
		waiting_times <- diff(windexi)
		lambda_init <- 1 / mean(waiting_times)
		mleHawkesi <- maxLik(logLik=logLikFun_Hawkes,start=c(lambda_init,alpha_init,beta_init),windex=windexi)

		for (j in (i+1):length(asset_list)) {
			if ((j <= length(asset_list)) & (j > i)) {
				print(paste("Hawkes 2D",i,j))

				sj <- loadjumps(asset_list[j],prefixes,param,path)
				windexj <- which(sj == TRUE) 
				waiting_times <- diff(windexj)
				lambda_init <- 1 / mean(waiting_times)
				mleHawkesj <- maxLik(logLik=logLikFun_Hawkes,start=c(lambda_init,alpha_init,beta_init),windex=windexj)

				cojprobemp  <- read.table(paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs"))[j,-(1:2)]
				cojprobempi <- read.table(paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs"))[i,-(1:2)]
				cojprobempj <- read.table(paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[j],"_obs"))[j,-(1:2)]

				# 2D Hawkes fit
				leni <- length(windexi)
				lenj <- length(windexj)
				lenmax <- max(leni,lenj)	
				windex2D <- rbind(rep(0,lenmax),rep(0,lenmax))
				windex2D[1,1:leni] <- windexi	
				windex2D[2,1:lenj] <- windexj	

				Hparami <- c(asin(mleHawkesi$estimate[1] / (0.5 * 0.005) - 1),asin(mleHawkesi$estimate[2] / 0.5 - 1),asin(0.01 / 0.5 - 1),asin(mleHawkesi$estimate[3] / 0.5 - 1),asin(0.01 / 0.5 - 1))
				Hparamj <- c(asin(mleHawkesj$estimate[1] / (0.5 * 0.005) - 1),asin(0.01 / 0.5 - 1),asin(mleHawkesj$estimate[2] / 0.5 - 1),asin(0.01 / 0.5 - 1),asin(mleHawkesj$estimate[3] / 0.5 - 1))
				Hparam0 <- rbind(Hparami,Hparamj)
				tmp <- mle2DHawkes_optimisation(maxsann,Hparam0,totlen,windex2D)

				if (!tmp$maximum) { print("Warning: optimisation for 2D Hawkes estimates was not performed correctly.") }
				Hparam <- tmp$Hparam
				stderr <- tmp$stderr
				print(c("Param:     ",Hparam))
				print(c("Std Error: ",stderr))
				print(c("t-Stat:    ",Hparam / stderr))
				print(c("p-value:   ",1 - pnorm(Hparam / stderr)))
				writeparam <- rbind(Hparam,stderr,Hparam / stderr,1 - pnorm(Hparam / stderr))
				wparamfile <- paste0(path$cojumps,"Hawkes_2D_CJ/",prefixes[period_counter],"_",asset_list[i],"-",asset_list[j],"_2DHest")
				# Check!! It maybe writes in an unexpected way.
				write.table(writeparam,file=wparamfile)

				# Simulate 2D Hawkes.
				cojprobi <- array(c(0),dim=c(MC,nwlen))
				cojprobj <- array(c(0),dim=c(MC,nwlen))
				cojprob  <- array(c(0),dim=c(MC,nwlen))
				for (mc in 1:MC) {
					if (mc %% 50 == 0) print(mc)
					new2DHawkes <- sim2DHawkes(totlen,Hparam,2)
					windexi <- new2DHawkes[1,which(new2DHawkes[1,] != 0)]
					windexj <- new2DHawkes[2,which(new2DHawkes[2,] != 0)]
					cojprobi[mc,] <- count_MJ(windexi,windowlengths,totlen)
					cojprobj[mc,] <- count_MJ(windexj,windowlengths,totlen)
					cojprob [mc,] <- count_CJ(windexi,windexj,windowlengths,totlen)
				}

				writefile <- paste0(path$cojumps,"Hawkes_2D_CJ/",prefixes[period_counter],"_",asset_list[i],"-",asset_list[j],"_2DHsim")
				MC_confidence_levels(cojprob ,nwlen,writefile)
				MC_confidence_levels(cojprobi,nwlen,writefile)
				MC_confidence_levels(cojprobj,nwlen,writefile)
			}
		}
	}
}


mle2DHawkes_optimisation <- function(maxsann,Hparam0,totlen,windex2D) {
	logLikelihood <- 0
	maximum <- FALSE
	cnt <- 1	# counter for the while cycle
	while (!maximum & cnt <= maxsann) {
		print(paste("sann:",cnt))
		BOX <- TRUE
		mle2DHawkes <- optim(Hparam0,logLikFun_2DHawkes,gr=NULL,method="SANN",control=list(maxit=1000,temp=20,fnscale=-1),BOX=BOX,T=totlen,t=windex2D)
		if (mle2DHawkes$value > logLikelihood) {
			lambda <- 0.005 * 0.5 * (1 + sin(mle2DHawkes$par[1:2]))
			alpha  <- 0.5 * (1 + sin(mle2DHawkes$par[3:6]))
			beta   <- 0.5 * (1 + sin(mle2DHawkes$par[7:10]))
			Hparami <- c(lambda[1],alpha[1],alpha[3],beta[1],beta[3])
			Hparamj <- c(lambda[2],alpha[2],alpha[4],beta[2],beta[4])
			tmp <- rbind(Hparami,Hparamj)
			BOX <- FALSE
			mle2DHawkesOptim <- optim(tmp,logLikFun_2DHawkes,method="CG",control=list(maxit=1000,fnscale=-1),hessian=TRUE,BOX=BOX,T=totlen,t=windex2D)
			tmp_mean <- mle2DHawkesOptim$par	
			if (prod(eigen(mle2DHawkesOptim$hessian)$values < 0) == 1) {
				logLikelihood <- mle2DHawkes$value
				Hparam <- tmp_mean
				covmatrix <- -solve(mle2DHawkesOptim$hessian)
				stderr <- matrix(sqrt(diag(covmatrix)),2,5)
				print(logLikelihood)
				maximum <- TRUE
			}
		}
		cnt <- cnt + 1
	}
	return(list(maximum=maximum,Hparam=Hparam,stderr=stderr))
}


logLikFun_2DHawkes <- function(pars,M=2,BOX,T,t) {
	if (BOX == TRUE) {
		lmb <- 0.005 * 0.5 * (1 + sin(pars[1:M]))
		aph <- 0.5 * (1 + sin(pars[(M+1):(M*M+M)]))
		bet <- 0.5 * (1 + sin(pars[(M*M+M+1):(2*M*M+M)]))
	} else {
		lmb <- pars[1:M]
		aph <- pars[(M+1):(M*M+M)]
		bet <- pars[(M*M+M+1):(2*M*M+M)]
		#print(c("ALPHA  is ",aph), quotes = FALSE)
		#print(c("BETA   is ",bet), quotes = FALSE)
		#print(c("LAMBDA is ",lmb), quotes = FALSE)
	}

	if (prod(lmb > 0) && prod(aph > 0) && prod(bet > 0)) {
		aph <- matrix(aph,M,M)
		bet <- matrix(bet,M,M)

		#print(lmb)
		#print(aph)
		#print(bet)

		logL  <- array(c(0), dim = c(M,1)) # Mx1 array along the components

		dt <- cbind(t[,1], rbind(diff(t[1,]),diff(t[2,]))) # M x nmax array with first differences of event times (included 0) through the columns
		dt[dt < 0] <- 0                                    # put to 0 the differences < 0, because they are due to the fact that the events of that type were finished
		t_all <- as.vector(t)                     # put all elements of the bi-dimensional t in a vector
		t_all <- t_all[t_all != 0]  # delete 0-elements

		for (m in 1:M) {
			nl <- sum(t[m,]>0)
			# Computation of the auxiliary and recursive variable R[m,n](k)
			# m is fixed, n and k vary
			R <- matrix(rep(0,M * nl),M,nl)  # R(0) = 0 (M x nl array)
			# R[m,n](1)
			if (m > 1) {
				for (n in 1:(m-1)) {
					tn1 <- t[n,(t[n,]<t[m,1])]
					tn1 <- tn1[tn1 != 0]
					R[n,1] <- sum(exp(-bet[m,n] * (t[m,1] - tn1)))
				}
			}
			R[m,1] <- exp(-bet[m,m] * dt[m,1])
			if (m < M) {
				for (n in (m+1):M) {
					tn1 <- t[n,(t[n,]<t[m,1])]
					tn1 <- tn1[tn1 != 0]
					R[n,1] <- sum(exp(-bet[m,n] * (t[m,1] - tn1)))
				}
			}
			# Recursive computation of R^{m,n} 
			for (k in 2:nl) {
				if (m > 1) {
					for (n in 1:(m-1)) {
						u <- (t[n,] <  t[m,k])
						l <- (t[n,] >= t[m,k-1])
						tnk <- t[n,((u+l)==2)]
						R[n,k] <- exp(-bet[m,n] * dt[m,k]) * R[n,(k-1)] + sum(exp(-bet[m,n] * (t[m,k] - tnk)))
					}
				}
				R[m,k] <- exp(-bet[m,m] * dt[m,k]) * (1 + R[m,(k-1)])
				if (m < M) {
					for (n in (m+1):M) {
						u <- (t[n,] <  t[m,k])
						l <- (t[n,] >= t[m,(k-1)])
						tnk <- t[n,((u+l)==2)]
						R[n,k] <- exp(-bet[m,n] * dt[m,k]) * R[n,(k-1)] + sum(exp(-bet[m,n] * (t[m,k] - tnk)))
					}
				}
			}

			# Computation of the log-likelihood for each component m:
			aux <- rep(0,M)
			for (n in 1:M) {
				aux[n] <- sum(1 - exp(-bet[m,n] * (T - t_all)))
			}
			logL[m] <- T * (1 - lmb[m]) - sum(aph[m,] / bet[m,] * aux) + sum(log(lmb[m] + aph[m,] %*% R))
		}	

		logL <- sum(logL) 
	} else {
		logL <- 0
	}
	logL
}


I2 <- function(pars,t_events,s,k,M) {
	# This function computes the total intensity up to the kth component of a two-dimensional Hawkes process
	# at time t, with events occurring in t_events.
	# INPUT:
	# pars = [lmb0(1), aph(1,1), aph(1,2), bet(1,1), bet(1,2);
	#         lmb0(2), aph(2,1), aph(2,2), bet(2,1), bet(2,2)];
	# t_events =  2xn_events array of event times of the process, where the
	#             first row contains the occurring times of events of type 1 and 0 when the
	#             events are of type 2; analogously for the second row
	# t = time at which the intensity has to be computed
	# k = component up to which compute the intensity (either 0, 1 or 2).

	lmb <- pars[,1]
	aph <- pars[,2:(1+M)]
	bet <- pars[,(2+M):(1+2*M)]

	lk <- matrix(rep(0,k),k,1) #kx1 array along the components

	# Computation for each component
	if (k > 0) {
		for (m in 1:k) {
			aux <- rep(0,M)
			for (n in 1:M) {
				past_events <- t_events[n,(t_events[n,] < s)]
				past_events <- past_events[past_events > 0]
				aux[n] <- sum(exp(-bet[m,n] * (s - past_events)))
			}
			lk[m] <- lmb[m] + sum(aph[m,] * aux)
		}
	}
	I2 <- sum(lk) # if k=0 -> Ik=0
	I2
}


sim2DHawkes <- function(T,pars,M) {
	# sim2DHawkes simulates a M-dimensional Hawkes process with P=1 and
	# parameters given in input, on the time interval [0,T]
	# INPUT:
	# T = final time for the simulation
	# pars = [lmb(1), aph(1,1), aph(1,2), bet(1,1), bet(1,2);
	#         lmb(2), aph(2,1), aph(2,2), bet(2,1), bet(2,2)]
	# M = dimension of the process
	# OUTPUT:
	# t_events =  2xn_events array of event times of the process, where the
	#             first row contains the occurring times of events of type 1 and 0 when the
	#             events are of type 2; analogously for the second row
	# tstar = vector of times of changes in value of I_max
	# I_max = maximum intensity, constant in each interval [ti,t_{i+1}]

	# Allocation
	lmb <- pars[,1]
	aph <- pars[,2:(M+1)]

	I_max    <- rep(0,12 * T)
	tstar    <- rep(0,12 * T)
	t_events <- matrix(rep(0, M * 10 * T),M,10 * T)

	# First step
	n <- 0  
	istar <- 1
	Istar <- sum(lmb)
	tstar[istar] <- 0
	I_max[istar] <- Istar
	U <- runif(1)
	s <- - 1 / Istar * log(U)  # generate the 1st event time from a Exp(Istar) dist.
	D <- runif(1)
	j <- M
	while ((j > 0) && (D <= sum(lmb[1:j]) / Istar)) {
		j <- j - 1
	}
	j <- j + 1

	# General routine
	while (s <= T) {
		n <- n + 1  # occurrence of the n-th event
		t_events[j,n] <- s  # the event is of type j; in the other rows we leave 0
		istar <- istar + 1
		if (n == 1) {
			tmp <- matrix(t_events[,1],2,1)
		} else {
			tmp <- t_events[,1:n]
		}
		Istar <- I2(pars,tmp,s,M,M) + sum(aph[,j]) # in the first step it's the sum of lmb
		I_max[istar] <- Istar
		tstar[istar] <- s
		U <- runif(1)
		s <- s - 1 / Istar * log(U)
		while (s <= T) {
			D <- runif(1)
			j <- M
			while ((j >= 0) && (D <= I2(pars,tmp,s,j,M) / Istar)) { # if D > I(pars,t_events(:,1:n),s,M)/Istar -> the while is not entered and J=M
				j <- j - 1
			}
			j <- j + 1
			if (j > M) {
				istar <- istar + 1
				Istar <- I2(pars,tmp,s,M,M)
				I_max[istar] <- Istar
				tstar[istar] <- s
				U <- runif(1)
				s <- s - 1 / Istar * log(U)
			} else {
				break
			}
		}
	}
	t_events <- round(t_events[,1:n])
	t_events
}	


MC_confidence_levels <- function(cojprob,nwlen,writefile) {
	mnmc <- colMeans(cojprob)

	lw99qmc <- rep(0,nwlen)
	up99qmc <- rep(0,nwlen)
	lw95qmc <- rep(0,nwlen)
	up95qmc <- rep(0,nwlen)
	for (wl in 1:nwlen) {
		tmp <- quantile(cojprob[,wl],probs=c(0.005 / nwlen,1 - 0.005 / nwlen))
		lw99qmc[wl] <- tmp[1]
		up99qmc[wl] <- tmp[2]
		tmp <- quantile(cojprob[,wl],probs=c(0.025 / nwlen,1 - 0.025 / nwlen))
		lw95qmc[wl] <- tmp[1]
		up95qmc[wl] <- tmp[2]
	}

	write(mnmc   ,file=writefile,append=TRUE,ncolumns=nwlen)
	write(lw99qmc,file=writefile,append=TRUE,ncolumns=nwlen)
	write(up99qmc,file=writefile,append=TRUE,ncolumns=nwlen)
	write(lw95qmc,file=writefile,append=TRUE,ncolumns=nwlen)
	write(up95qmc,file=writefile,append=TRUE,ncolumns=nwlen)
}


plot_CJ_against_2DHawkes <- function(asset_list,prefixes,path,windowlengths) {
	pdf(paste0(path$cojumps,"Hawkes_2D_CJ/",prefixes[period_counter],"_CJ_2DH.pdf"),width=8,height=6)
		ylab <- "CJ frequency"
		for (i in 1:length(asset_list)) {
			for (j in (i+1):length(asset_list)) {
				if ((j <= length(asset_list)) & (j > i)) {
					inputfile <- paste0(path$cojumps,"Hawkes_2D_CJ/",prefixes[period_counter],"_",asset_list[i],"-",asset_list[j],"_2DHsim")

					CJp     <- read.table(paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs"))[j,-(1:2)]
					MCres <- read.table(inputfile)[1:5,]
					plotMJCJ_withMCcl(windowlengths,CJp,MCres,ylab,plottitle=paste("CJ probability of",asset_list[i],"and",asset_list[j]))
				}
			}
		}
	dev.off()
}


plot_MJ_against_2DHawkes <- function(asset_list,prefixes,path,windowlengths) {
	pdf(paste0(path$cojumps,"Hawkes_2D_CJ/",prefixes[period_counter],"_MJ_2DH.pdf"),width=8,height=6)
		ylab <- "MJ frequency"
		for (i in 1:length(asset_list)) {
			for (j in (i+1):length(asset_list)) {
				if ((j <= length(asset_list)) & (j > i)) {
					inputfile <- paste0(path$cojumps,"Hawkes_2D_CJ/",prefixes[period_counter],"_",asset_list[i],"-",asset_list[j],"_2DHsim")

					MJp <- read.table(paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs"))[i,-(1:2)]
					MCres <- read.table(inputfile)[6:10,]
					plotMJCJ_withMCcl(windowlengths,MJp,MCres,ylab,plottitle=paste("MJ probability of",asset_list[i]))

					MJp <- read.table(paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[j],"_obs"))[j,-(1:2)]
					MCres <- read.table(inputfile)[11:15,]
					plotMJCJ_withMCcl(windowlengths,MJp,MCres,ylab,plottitle=paste("MJ probability of",asset_list[j]))
				}
			}
		}
	dev.off()
}


plotMJCJ_withMCcl <- function(windowlengths,MJCJp,mcmn_and_mcq,ylab,plottitle) {
	mnmc    <- mcmn_and_mcq[1,]
	lw99qmc <- mcmn_and_mcq[2,]
	up99qmc <- mcmn_and_mcq[3,]
	lw95qmc <- mcmn_and_mcq[4,]
	up95qmc <- mcmn_and_mcq[5,]

	ymax <- max(up99qmc,MJCJp)
	ymin <- min(lw99qmc,MJCJp)
	plot (windowlengths,mnmc   ,type="o",col="lightblue",ylim=c(ymin,ymax),ylab=ylab,xlab="window width",main=plottitle)
	lines(windowlengths,lw99qmc,type="o",col="red")
	lines(windowlengths,up99qmc,type="o",col="red")
	lines(windowlengths,lw95qmc,type="o",col="orange")
	lines(windowlengths,up95qmc,type="o",col="orange")
	lines(windowlengths,MJCJp  ,type="o",col="blue")
	legend(x="topleft",legend=c("observed","MC mean","MC 99% c.l.","MC 95% c.l."),col=c("blue","lightblue","red","orange"),lty="solid",pch="o",inset=0.04)
}

##### end of Hawkes_2D_CJ

##### start of Hawkes_1F_MJ_CJ

Hawkes_1F_MJ_CJ <- function(asset_list,prefixes,param,path,totlen,windowlengths) {
	J <- find_critical_J(asset_list,prefixes,param,path)
	factor_Hawkes_idiosync_Hawkes(asset_list,prefixes,path,totlen,windowlengths,J)
	plot_1F(asset_list,prefixes,path,windowlengths,J)
}


find_critical_J <- function(asset_list,prefixes,param,path) {
	all_jumps <- all_jumps_binary(asset_list,prefixes,param,path)
	obscoj <- hist(rowSums(all_jumps),breaks=0:(length(asset_list) + 1) - 0.5)$counts

	nshuffle <- 100
	signif <- 0.01
	rndcojmatrix <- array(0,dim=c(nshuffle,length(asset_list)+1))
	print(paste0("Shuffling jump vectors (",nshuffle," shuffles)..."))
	for (i in 1:nshuffle) {
		if (i %% 20 == 0) print(i)
		for (j in 1:ncol(all_jumps)) {
			all_jumps[,j] <- xts(coredata(all_jumps[,j]),sample(index(all_jumps)))
		}
		rndcojmatrix[i,] <- hist(rowSums(all_jumps),breaks=0:(length(asset_list) + 1) - 0.5)$counts
	}
	rndcoj <- colMeans(rndcojmatrix)
	J <- first(which(rndcoj / obscoj < signif)) - 1
	write(c(prefixes[period_counter],length(asset_list),J),file=paste0(path$cojumps,"Hawkes_1F/J"),append=TRUE,ncolumns=3)
	return(J)
}


all_jumps_binary <- function(asset_list,prefixes,param,path) {
	allj <- xts()
	for (i in 1:length(asset_list)) {
		s <- loadjumps(asset_list[i],prefixes,param,path)
		allj <- cbind(allj,jumps_binary_series(s))
	}
	allj
}


# jumps_binary_series transforms a boolean series of jumps (with TRUE, FALSE and NA) into a binary series, with 1 in correspondence of jumps.
jumps_binary_series <- function(jumps_boolean_xts) {
	j <- xts(rep(0,length(jumps_boolean_xts)),index(jumps_boolean_xts))
	j[which(jumps_boolean_xts == TRUE)] <- 1
	return(j)
}


factor_Hawkes_idiosync_Hawkes <- function(asset_list,prefixes,path,totlen,windowlengths,J) {
	nwlen <- length(windowlengths)
	nasset <- length(asset_list)
	mleparam <- array(0,dim=c(nasset,3))

	allju <- readRDS(paste0(path$aggregate,prefixes[period_counter],"_jumps.RDS"))
	## remove VOD(afone) (Ad hoc adjustment for first run. Make it smoother...)
	#if ("VOD" %in% colnames(allju)) {
	#	vodcol <- which(colnames(allju) == "VOD")
	#	allju <- allju[,-vodcol]
	#}

	# Define the factor and estimate its Hawkes parameters
	coj <- rowSums(abs(allju))
	Findex <- which(coj >= J)
	Flambda <- 1 / mean(diff(Findex))
	Falpha  <- 0.015
	Fbeta   <- 0.2
	mleHawkes <- maxLik(logLik=logLikFun_Hawkes,start=c(Flambda,Falpha,Fbeta),windex=Findex)
	Hparams <- summary(mleHawkes)$estimate
	Hestim  <- c(Hparams[,1],Hparams[,4])
	write(Hestim,file=paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_1FHIH_estim_and_pval"),append=TRUE,ncolumns=6)

	# Idiosyncratic Hawkes
	ilambda <- array(0,dim=c(nasset,1))
	ialpha  <- array(0,dim=c(nasset,1))
	ibeta   <- array(0,dim=c(nasset,1))
	
	# Number of jumps of single assets Ni, jumps in common with the factor NiF, and asset's Pi
	Ni  <- array(0,dim=c(nasset,1))
	NiF <- array(0,dim=c(nasset,1))
	Pi  <- array(0,dim=c(nasset,1))

	alpha0 <- 0.015
	beta0  <- 0.2
	print("Estimating Hawkes processes on asset jump series...")
	for (i in 1:nasset) {
		if (i %% 10 == 0) print(i)

		windexi <- which(allju[,i] != 0)
		Ni[i]   <- length(windexi)
		NiF[i]  <- length(intersect(windexi,Findex))
		Pi[i]   <- NiF[i] / (totlen * Flambda) * (1 - Falpha / Fbeta)
		# Flambda * totlen / (1 - Falpha / Fbeta) gives the expected number of the factor's jumps

		windex <- setdiff(windexi,Findex)
		lambda0 <- 1 / mean(diff(windex))
		imleHawkes <- maxLik(logLik=logLikFun_Hawkes,start=c(lambda0,alpha0,beta0),windex=windex)
		ilambda[i] <- imleHawkes$estimate[1]
		ialpha[i]  <- imleHawkes$estimate[2]
		ibeta[i]   <- imleHawkes$estimate[3]
		Hparams <- summary(imleHawkes)$estimate
		Hestim  <- c(Hparams[,1],Hparams[,4])
		write(Hestim,file=paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_1FHIH_estim_and_pval"),append=TRUE,ncolumns=6)
	}
	write.table(Pi,file=paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_Pi"),col.names=FALSE)

	simulate_1FHIH(asset_list,prefixes,path,mleHawkes,MC,totlen,windowlengths,ilambda,ialpha,ibeta,Pi,J)
}


simulate_1FHIH <- function(asset_list,prefixes,path,mleHawkes,MC,totlen,windowlengths,ilambda,ialpha,ibeta,Pi,J) {
	nasset <- length(asset_list)
	nwlen <- length(windowlengths)
	
	# One factor Hawkes
	lambda <- mleHawkes$estimate[1]
	alpha  <- mleHawkes$estimate[2]
	beta   <- mleHawkes$estimate[3]

	# Monte Carlo
	CJprobmn <- array(0,dim=c(nasset * (nasset - 1) / 2,nwlen))
	CJprobsq <- array(0,dim=c(nasset * (nasset - 1) / 2,nwlen))
	MJprobmn <- array(0,dim=c(nasset,nwlen))
	MJprobsq <- array(0,dim=c(nasset,nwlen))
	print("One-factor Hawkes Idiosyncratic Hawkes Monte Carlo simulation begins...")
	for (mc in 1:MC) {
		if (mc %% 50 == 0) print(mc)
		newHawkes <- simHawkes(totlen,c(lambda,alpha,beta))
		njuF <- length(newHawkes)

		# Big matrix of simulated jumps
		jumps <- array(0,dim=c(nasset,totlen))
		for (i in 1:nasset) {
			# Asset i jumps with the factor with probability Pi...
			u <- runif(njuF)
			jumps[i,newHawkes[which(u <= Pi[i])]] <- 1

			# ...and also according to its idiosyncratic Hawkes.
			inewHawkes <- simHawkes(totlen,c(ilambda[i],ialpha[i],ibeta[i]))
			jumps[i,inewHawkes] <- 1
		}

# The following command, together with the gen_jumps function, is alternative to the for loop above. Although more elegant, it performs slightly worse.
#		jumps <- do.call("rbind",lapply(1:nasset,gen_jumps,newHawkes,totlen,njuF,Pi,ilambda,ialpha,ibeta))
		
		# Find simulated jumps for all assets.
		windex <- apply(jumps,1,function(x) which(x == 1))	# It's a list!

		# Count MJ
		tmp <- do.call("rbind",lapply(windex,count_MJ,windowlengths,totlen))
		MJprobmn <- MJprobmn + tmp
		MJprobsq <- MJprobsq + tmp^2
		
		# Count CJ
		idxvec <- rbind(rep(1:(nasset-1),(nasset-1):1),unlist(lapply(2:nasset,function(x) x:nasset)))
		tmp <- t(apply(idxvec,2,function(x) count_CJ(windex[[x[1]]],windex[[x[2]]],windowlengths,totlen)))
		CJprobmn <- CJprobmn + tmp
		CJprobsq <- CJprobsq + tmp^2
	}

	MJmn <- MJprobmn / MC
	MJsq <- MJprobsq / MC
	MJsd <- sqrt(MJsq - MJmn^2)

	CJmn <- CJprobmn / MC
	CJsq <- CJprobsq / MC
	CJsd <- sqrt(CJsq - CJmn^2)

	outputfile <- paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_MC_MJmn_J",J)
	indexedmatrix <- cbind(1:nrow(MJmn),1:nrow(MJmn),MJmn)
	write.table(indexedmatrix,file=outputfile,col.names=FALSE,row.names=FALSE)
	outputfile <- paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_MC_MJsd_J",J)
	indexedmatrix <- cbind(1:nrow(MJsd),1:nrow(MJsd),MJsd)
	write.table(indexedmatrix,file=outputfile,col.names=FALSE,row.names=FALSE)

	outputfile <- paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_MC_CJmn_J",J)
	indexedmatrix <- cbind(rep(1:(nasset-1),(nasset-1):1),unlist(lapply(2:nasset,function(x) x:nasset)),CJmn)
	write.table(indexedmatrix,file=outputfile,col.names=FALSE,row.names=FALSE)
	outputfile <- paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_MC_CJsd_J",J)
	indexedmatrix <- cbind(rep(1:(nasset-1),(nasset-1):1),unlist(lapply(2:nasset,function(x) x:nasset)),CJsd)
	write.table(indexedmatrix,file=outputfile,col.names=FALSE,row.names=FALSE)
}


#gen_jumps <- function(assind,newHawkes,totlen,njuF,Pi,ilambda,ialpha,ibeta) {
#	jumps <- array(0,dim=totlen)
#
#	# Asset i jumps with the factor with probability Pi...
#	u <- runif(njuF)
#	jumps[newHawkes[which(u <= Pi[assind])]] <- 1
#
#	# ...and also according to its idiosyncratic Hawkes.
#	inewHawkes <- simHawkes(totlen,c(ilambda[assind],ialpha[assind],ibeta[assind]))
#	jumps[inewHawkes] <- 1
#	jumps
#}


plot_1F <- function(asset_list,prefixes,path,windowlengths,J) {
	print("Testing and plotting against 1FHIH model...")
	pdf(paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_1FHIH_J",J,".pdf"),width=8,height=6)
		nasset <- length(asset_list)
		nwlen <- length(windowlengths)
		sl1 <- 0.01; sl1bonf <- sl1 / nwlen
		nsd1 <- qnorm(1-sl1bonf/2)
		sl2 <- 0.05; sl2bonf <- sl2 / nwlen
		nsd2 <- qnorm(1-sl2bonf/2)

		MJmn <- read.table(paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_MC_MJmn_J",J))
		MJsd <- read.table(paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_MC_MJsd_J",J))
		CJmn <- read.table(paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_MC_CJmn_J",J))
		CJsd <- read.table(paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_MC_CJsd_J",J))
		for (i in 1:nasset) {
			#if (i == 2) break	# To process just the first asset.

			empiricalfile <- paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs")
			MJp <- read.table(empiricalfile)[i,][-(1:2)]

			mnmc <- MJmn[i,]
			if (mnmc[1:2] != c(i,i)) stop("Wrong index read!")
			sdmc <- MJsd[i,]
			if (sdmc[1:2] != c(i,i)) stop("Wrong index read!")
			modelok <- test_and_plot_emp_vs_MC1F(windowlengths,MJp,nsd1,nsd2,mnmc[-(1:2)],sdmc[-(1:2)],ylab="MJ frequency",plottitle=paste("MJ probability of",asset_list[i]))
			write(c(i,i,modelok),file=paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_1FHIH_modelok_J",J),append=TRUE,ncolumns=4)

			for (j in (i+1):nasset) {
				if (i != nasset) {
					empiricalfile <- paste0(path$cojumps,"observed/",prefixes[period_counter],"_",asset_list[i],"_obs")
					CJp <- read.table(empiricalfile)[j,][-(1:2)]

					row <- intersect(which(CJmn[,1] == i),which(CJmn[,2] == j))		# Cambiare!
					mnmc <- CJmn[row,]
					#if (mnmc[1:2] != c(i,j)) stop("Wrong index read!")
					row <- intersect(which(CJsd[,1] == i),which(CJsd[,2] == j))		# Cambiare!
					sdmc <- CJsd[row,]
					#if (sdmc[1:2] != c(i,j)) stop("Wrong index read!")
					modelok <- test_and_plot_emp_vs_MC1F(windowlengths,CJp,nsd1,nsd2,mnmc[-(1:2)],sdmc[-(1:2)],ylab="CJ frequency",plottitle=paste("CJ probability of",asset_list[i],"and",asset_list[j]))
					write(c(i,j,modelok),file=paste0(path$cojumps,"Hawkes_1F/",prefixes[period_counter],"_1FHIH_modelok_J",J),append=TRUE,ncolumns=4)
				}
			}
		}
	dev.off()
}


test_and_plot_emp_vs_MC1F <- function(windowlengths,MJCJp,nsd1,nsd2,mnmc,sdmc,ylab,plottitle) {
	# Confidence levels
	up99   <- mnmc + nsd1 * sdmc
	down99 <- mnmc - nsd1 * sdmc
	down99 <- 0.5 * (1 + sign(down99)) * down99
	up95   <- mnmc + nsd2 * sdmc
	down95 <- mnmc - nsd2 * sdmc
	down95 <- 0.5 * (1 + sign(down95)) * down95

	# Plot
	ymax <- max(up99,MJCJp)
	ymin <- min(down99,MJCJp)
	plot (windowlengths,mnmc  ,type="o",col="lightblue",ylim=c(ymin,ymax),ylab=ylab,xlab="window width",main=plottitle)
	lines(windowlengths,down99,type="o",col="red")
	lines(windowlengths,up99  ,type="o",col="red")
	lines(windowlengths,down95,type="o",col="orange")
	lines(windowlengths,up95  ,type="o",col="orange")
	lines(windowlengths,MJCJp ,type="o",col="blue")
	legend(x="topleft",legend=c("observed","MC mean","MC 95% c. l.","MC 99% c. l."),col=c("blue","lightblue","orange","red"),lty="solid",pch="o",inset=0.04)

	# Test
	in99 <- prod(down99 <= MJCJp & MJCJp <= up99)
	in95 <- prod(down95 <= MJCJp & MJCJp <= up95)
	return(c(in99,in95))
}

##### end of Hawkes_1F_MJ_CJ
